import logging
import re
from abc import ABC, abstractmethod
from enum import Enum

import pandas as pd

'''
This class allows for different execution styles. 
In particular, it allows for dependent queries, where the result of the first query is used as a parameter in the second query.
'''

_logger = logging.getLogger(__name__)


class QueryExecutor(ABC):
    @abstractmethod
    def execute(self, cursor):
        pass

    @abstractmethod
    def queries(self):
        pass

    def _replace_placeholders(self, query):
        # replace :xyz with %(xyz)s using regex
        return re.sub(r':(\w+)', r'%(\1)s', query)


class SimpleQueryExecutor(QueryExecutor):
    def __init__(self, line_number, operation_type, table_name, query, params, context):
        self.line_number = line_number
        self.operation_type = operation_type
        self.table_name = table_name
        self.query = query
        self.params = params
        self.context = context

    def queries(self):
        return [(self.query, self.params)]

    def execute(self, cursor):
        # replace ':' style place holders with '%' style
        psycopg_query = self._replace_placeholders(self.query)
        # replace NA values with None in params dictionary
        params_with_none = {k: None if pd.isna(v) else v for k, v in self.params.items()}

        try:
            # execute query
            cursor.execute(psycopg_query, params_with_none)
        except Exception as e:
            error = str(e)
            return ExecutionResult(self.line_number, self.operation_type, False, 0, self.table_name, self.query, self.params, self.context, error=error)

        # Get the number of affected rows
        rowcount = cursor.rowcount

        # verify row was inserted
        if rowcount == 0:
            error = 'No row was inserted'
            return ExecutionResult(self.line_number, self.operation_type, False, rowcount, self.table_name, self.query, self.params, self.context, error=error)

        # verify no more than one row was inserted
        if rowcount > 1:
            # we must not commit the transaction
            error = "More than one row was inserted, do not commit."
            return ExecutionResult(self.line_number, self.operation_type, False, rowcount, self.table_name, self.query, self.params, self.context, error, True)

        result = ExecutionResult(self.line_number, self.operation_type, True, rowcount, self.table_name, self.query, self.params, self.context)
        return result


class DependentQueryExecutor(QueryExecutor):
    def __init__(self, initial_query, dependent_query):
        self.query = initial_query[0]
        self.params = initial_query[1]
        self.dependent_query = dependent_query

    def queries(self):
        return [(self.query, self.params), self.dependent_query]

    def execute(self, cursor):
        # replace ':' style place holders with '%' style
        query_0 = self._replace_placeholders(self.query)

        # replace NA values with None in params dictionary
        params_0 = {k: None if pd.isna(v) else v for k, v in self.params.items()}

        cursor.execute(query_0, params_0)
        result = cursor.fetchone()
        rowcount = cursor.rowcount

        # verify row was inserted
        if rowcount == 0:
            _logger.warning("No row was inserted. Query: %s, Params: %s" % (query_0, params_0))
            return (rowcount, query_0, params_0)

        # verify no more than one row was inserted
        if rowcount > 1:
            # raise exception, bec/ we must not commit the transaction
            raise ValueError("More than one row was inserted. Inserts: %s, Query: %s, Params: %s" % (rowcount, query_0, params_0))

        # replace ':' style place holders with '%' style
        query_1 = self._replace_placeholders(self.dependent_query[0])
        params_1 = self.dependent_query[1]
        params_1['res_id'] = result[0]

        cursor.execute(query_1, params_1)
        result = (rowcount, query_0, params_0)

        return result


class OperationType(Enum):
    INSERT = "INSERT"
    UPDATE = "UPDATE"
    DELETE = "DELETE"

    def __str__(self):
        return self.value

    def __repr__(self):
        return "'" + self.value + "'"


class ExecutionResult:
    def __init__(self, line_number, operation_type, success, rowcount, table_name, query, params, context, error=None, block_commit=False):
        self.line_number = line_number
        if not isinstance(operation_type, OperationType):
            raise ValueError(f"operation_type must be an instance of OperationType Enum, not {type(operation_type)}")
        self.operation_type = operation_type
        self.success = success
        self.rowcount = rowcount
        self.table = table_name
        self.query = query
        self.params = params
        self.context = context
        self.error = error
        self.block_commit = block_commit

    def __str__(self):
        return f"Type: {self.operation_type}, Success: {self.success}, Rowcount: {self.rowcount}, Table: {self.table_name}, Query: {self.query}, Params: {self.params}"

    def to_dict(instance, execute):
        if execute:
            return {key: value for key, value in vars(instance).items() if value is not None and key != 'block_commit'}
        else:
            return {key: value for key, value in vars(instance).items() if value is not None and key not in ['block_commit', 'success', 'rowcount']}



"""
This class provides the main database service to list tables, create mappings, retrieve table contents and to post updates.

Author: Romke Jonker
Email: romke@rnadesign.net
"""
import importlib.util
import logging
import os
import re
import sys
from io import StringIO

import pandas as pd
import psycopg2
from pandas import DataFrame

from stimula.compiler.header_compiler import HeaderCompiler
from stimula.compiler.select_compiler import SelectCompiler
from stimula.compiler.types_compiler import TypesCompiler
from stimula.header.csv_header_parser import HeaderParser
from stimula.header.header_lexer import HeaderLexer
from stimula.header.header_merger import HeaderMerger
from stimula.header.odoo_header_parser import OdooHeaderParser
from .context import cnx_context, get_metadata
from .diff_to_sql import DiffToSql
from .query_executor import ExecutionResult, OperationType

_logger = logging.getLogger(__name__)


class DB:
    def __init__(self):
        self._lexer = HeaderLexer()
        self._diff_to_sql = DiffToSql()

    def get_tables(self, filter=None):

        cr = cnx_context.cr

        # query to get all tables, excluding views
        query = "SELECT table_name FROM information_schema.tables WHERE table_schema = 'public' AND table_type = 'BASE TABLE'"

        cr.execute(query)

        # fetch all rows
        all_tables = cr.fetchall()

        if filter is None:
            # return all tables
            unsorted = [t for t in all_tables]
        else:
            # use filter as a regular expression
            regex = re.compile(filter)
            unsorted = [t for t in all_tables if regex.search(t[0])]

        # get row counts
        with_counts = self.get_counts(unsorted)

        # return sorted(with_count)
        sorted_tables = sorted(with_counts, key=lambda e: e[0])

        # convert to dictionary for JSON serialization
        return [{'name': table_name, 'count': row_count} for table_name, row_count in sorted_tables]

    def get_counts(self, tables):
        cr = cnx_context.cr
        for table in tables:
            query = f"SELECT COUNT(*) FROM {table[0]}"
            cr.execute(query)
            row_count = cr.fetchone()[0]
            yield table[0], row_count

    def get_count(self, table_name, header, where_clause):
        # create metadata object from context
        metadata = get_metadata(cnx_context.cnx)

        if not header:
            # if header parameter is empty, then use the default header
            default_mapping = OdooHeaderParser(metadata, cnx_context.cr).parse(table_name)

            # remove columns that are not enabled
            mapping = {**default_mapping, 'columns': [c for c in default_mapping['columns'] if c.get('enabled', False)]}

        else:
            # parse header to build syntax tree
            mapping = HeaderParser(metadata, table_name).parse_csv(header)

        query = SelectCompiler().compile_count_query(mapping, where_clause)

        cr = cnx_context.cr

        # execute query
        cr.execute(query)

        # fetch result
        result = cr.fetchone()

        return result[0]

    def get_table(self, table_name, header, where_clause=None):
        # create metadata object from context
        metadata = get_metadata(cnx_context.cnx)

        if not header:
            # if header parameter is empty, then use the default header
            default_mapping = OdooHeaderParser(metadata, cnx_context.cr).parse(table_name)

            # remove columns that are not enabled
            mapping = {**default_mapping, 'columns': [c for c in default_mapping['columns'] if c.get('enabled', False)]}

        else:
            # parse header to build syntax tree
            mapping = HeaderParser(metadata, table_name).parse_csv(header)

        df = self._read_from_db(mapping, where_clause)

        return df, mapping

    def get_table_as_csv(self, table_name, header, where_clause, escapechar=''):
        # get table as dataframe
        df, mapping = self.get_table(table_name, header, where_clause)

        return self.convert_to_csv(df, mapping, escapechar)

    def convert_to_csv(self, df, mapping, escapechar=''):
        # need converters from types compiler
        column_names = HeaderCompiler().compile_list(mapping)
        column_types = TypesCompiler().compile(mapping, column_names)

        # get converters from column_types
        converters = column_types['write_csv_converters']

        # apply converters to dataframe
        for column in column_names:
            if column in converters:
                df[column] = df[column].apply(converters[column])

        # convert dataframe to csv
        return df.to_csv(index=False, escapechar=escapechar)

    def post_table_get_diff(self, table_name, header, where_clause, body, skiprows=0, insert=False, update=False, delete=False, execute=False, commit=False, deduplicate=False, post_script=None, context=None):
        # create diffs and sql
        diffs, sql = self._get_diffs_and_sql(table_name, header, where_clause, body, skiprows, insert, update, delete, deduplicate, post_script, context)
        # if execute
        if execute:
            # execute sql statements
            results_tuple = self._execute_sql(sql, commit)
            # concatenate results
            results = [result for results in results_tuple for result in results]
            # zip row counts with diffs statements
            for diff, result in zip(diffs, results):
                diff.append(result.rowcount)

        return diffs

    def post_table_get_sql(self, table_name, header, where_clause, body, skiprows=0, insert=False, update=False, delete=False, execute=False, commit=False, deduplicate=False, post_script=None, context=None):
        # create diffs and sql
        _, query_executors = self._get_diffs_and_sql(table_name, header, where_clause, body, skiprows, insert, update, delete, deduplicate, post_script, context)

        if execute:
            # execute sql statements
            sqls = self._execute_sql(query_executors, commit)
        else:
            sqls = [ExecutionResult(qe.line_number, qe.operation_type, False, 0, table_name, qe.query, qe.params, context) for qe in query_executors]

        # convert sql to dataframe
        return self._convert_to_df(sqls, execute)

    def post_table_get_full_report(self, table_name, header, where_clause, body, skiprows=0, insert=False, update=False, delete=False, execute=False, commit=False, deduplicate=False,
                                   post_script=None, context=None):
        # create diffs and sql
        diff, query_executors = self._get_diffs_and_sql(table_name, header, where_clause, body, skiprows, insert, update, delete, deduplicate, post_script, context)

        if execute:
            # execute sql statements
            execution_results = self._execute_sql(query_executors, commit)
        else:
            execution_results = [ExecutionResult(qe.line_number, qe.operation_type, False, 0, table_name, qe.query, qe.params, context) for qe in query_executors]

        # create full report
        return self._create_post_report(execution_results, execute, commit)

    def post_multiple_tables_get_full_report(self, table_names, header, where_clause, contents, skiprows=0, insert=False, update=False, delete=False, execute=False, commit=False, deduplicate=False,
                                             post_script=None, context=None):
        assert len(table_names) == len(contents), f"Provide exactly one file for each table name, so {len(table_names)}, not {len(contents)}"
        assert header is None, "Header must be None when posting multiple tables"
        assert skiprows ==1, "Skiprows must be 1 when posting multiple tables"
        assert deduplicate == False, "Deduplicate must be False when posting multiple tables"
        assert post_script is None, "Post script must be None when posting multiple tables"
        assert context is not None and len(context) == len(table_names), "Provide exactly one context for each table name, not %s" % len(context or [])

        query_executors = []

        # Iterate over tables here.
        for table_name, file_context, content in zip(table_names, context, contents):
            # decode binary content
            text_content = content.decode('utf-8')

            # get header from first line
            header = text_content.split('\n', 1)[0]

            # create diffs and sql
            _, qe = self._get_diffs_and_sql(table_name, header, where_clause, text_content, skiprows, insert, update, delete, deduplicate, post_script, file_context)
            query_executors.extend(qe)

        if execute:
            # execute sql statements
            execution_results = self._execute_sql(query_executors, commit)
        else:
            execution_results = [ExecutionResult(qe.line_number, qe.operation_type, False, 0, table_name, qe.query, qe.params, file_context) for qe in query_executors]

        # create full report
        return self._create_post_report(execution_results, execute, commit)

    def _create_post_report(self, execution_results, execute, commit):
        insert = len([er for er in execution_results if er.operation_type == OperationType.INSERT])
        update = len([er for er in execution_results if er.operation_type == OperationType.UPDATE])
        delete = len([er for er in execution_results if er.operation_type == OperationType.DELETE])

        found = {'insert': insert, 'update': update, 'delete': delete}

        summary = {'found': found, 'execute': execute, 'commit': commit}

        result = {'summary': summary}

        # only set success & failed if execute is True
        if execute:
            # summarize successful operations
            summary['success'] = {'insert': len([er for er in execution_results if er.operation_type == OperationType.INSERT and er.success]),
                       'update': len([er for er in execution_results if er.operation_type == OperationType.UPDATE and er.success]),
                       'delete': len([er for er in execution_results if er.operation_type == OperationType.DELETE and er.success])}

            # summarize failed operations
            summary['failed'] = {'insert': len([er for er in execution_results if er.operation_type == OperationType.INSERT and not er.success]),
                      'update': len([er for er in execution_results if er.operation_type == OperationType.UPDATE and not er.success]),
                      'delete': len([er for er in execution_results if er.operation_type == OperationType.DELETE and not er.success])}

        # list execution results
        rows = [er.to_dict(execute) for er in execution_results]
        result['rows'] = rows

        return result

    def _convert_to_df(self, sqls, showResult):
        # create empty pandas dataframe. First column contains sql
        if showResult:
            result = pd.DataFrame(columns=['rows', 'sql'])
        else:
            result = pd.DataFrame(columns=['sql'])

        # iterate sql statements
        index = 0
        for er in sqls:
            # create dictionary with query and value_dict
            value_dict = {**er.params, 'sql': er.query}
            if showResult:
                value_dict = {**value_dict, 'rows': er.rowcount}
            # append a new row to the bottom result with values from value dictionary using concat
            result = pd.concat([result, pd.DataFrame(value_dict, index=[index])], ignore_index=True)
            index += 1

        # force pandas to not convert int columns to float if they contain NaNs
        result = result.convert_dtypes()

        # return dataframe
        return result

    def post_table_get_summary(self, table_name, header, where_clause, body, skiprows=0, insert=False, update=False, delete=False, execute=False, commit=False, deduplicate=False):
        with self._database_session() as (engine, metadata):
            # parse header to build syntax tree
            diffs, sql = self._get_diffs_and_sql(engine, metadata, table_name, header, where_clause, body, skiprows, insert, update, delete)

            # create result object
            result = {}

            if execute:
                # execute sql statements
                results_tuple = self._execute_sql(engine, sql, commit)
                # summarize results
                # zip row counts with sql statements

            return result

    def _get_diffs_and_sql(self, table_name, header, where_clause, body, skiprows, insert, update, delete, deduplicate, post_script, context):
        # get cnx from context
        cnx = cnx_context.cnx

        # parse header to build syntax tree
        mapping = HeaderParser(get_metadata(cnx), table_name).parse_csv(header)

        # read dataframe from request first, so we can give feedback on errors in the request
        df_request = self._read_from_request(mapping, body, skiprows, deduplicate, post_script)

        # read dataframe from DB
        df_db = self._read_from_db(mapping, where_clause, set_index=True)

        # compare to obtain diff
        diffs = self._compare(df_request, df_db, insert, update, delete)

        # create sql statements and parameters
        sqls = self._diff_to_sql.diff_sql(mapping, diffs, context)

        return diffs, sqls

    def _read_from_request(self, mapping, body, skiprows, deduplicate=False, post_script=None):
        # get columns and unique columns.
        column_names = HeaderCompiler().compile_list(mapping, include_skip=True)
        index_columns = HeaderCompiler().compile_list_unique(mapping)
        column_types = TypesCompiler().compile(mapping, column_names, include_skip=True)

        # assert that at least one column header is not empty
        if not [c for c in column_names if c != '']:
            raise ValueError("At least one column header must not be empty")

        # read body as csv to count the number of columns in the body
        # there may be a more efficient way without having to read the csv twice.
        body_column_count = self._count_body_columns(body)

        # pad column names with empty columns if the body has more columns than the header
        if body_column_count > len(column_names):
            # add empty columns to the header
            column_names += [''] * (body_column_count - len(column_names))

        # replace empty column names with skip, skip1, skip2. This is because pandas requires column names to be unique
        non_empty_column_names = list(self._replace_empty_columns_with_skip(column_names))

        # find duplicate column names
        duplicate_column_names = self._find_duplicate_names(non_empty_column_names)

        # if there are duplicate column names, raise an exception
        if duplicate_column_names:
            raise ValueError(f"Duplicate column names are not supported: {', '.join(duplicate_column_names)}")

        # get list of columns to use in the output dataframe
        use_columns = [c for c in non_empty_column_names if c in column_names]

        # list names of columns with datetime64 or date type, because we need to parse them as datetime
        parse_dates = column_types.get('read_csv_parse_dates', {})

        # get converter dictionary
        converters = column_types.get('read_csv_converters', {})

        # get dtypes for read_csv
        dtype = column_types.get('read_csv_dtypes', {})

        # create initial column names to read the csv before padding
        initial_index_columns = [c for c in index_columns if c in non_empty_column_names[:body_column_count]]
        initial_names = non_empty_column_names[:body_column_count]
        initial_usecols = [c for c in use_columns if c in non_empty_column_names[:body_column_count]]

        # read csv from request body
        # treat '' as missing value, but treat NA as string
        df = pd.read_csv(
            StringIO(body),
            names=initial_names,
            index_col=initial_index_columns,
            skipinitialspace=True,
            skiprows=skiprows,
            dtype=dtype,
            # skip the empty column names that we renamed to skip, skip1, etc
            usecols=initial_usecols,
            parse_dates=parse_dates,
            converters=converters,
            na_values=[''],
            keep_default_na=False
        )

        df_padded = self._pad_dataframe_with_empty_columns(df, use_columns, index_columns, converters)

        # evaluate column expressions
        self._evaluate_expressions(df_padded, mapping, index_columns)

        # insert a column with line numbers
        df_padded.insert(0, '__line__', range(0, len(df)))

        # deduplicate if requested
        if deduplicate:
            df_padded = self._deduplicate(df_padded, index_columns)

        # verify that there are no duplicate index values
        if df_padded.index.has_duplicates:
            # find duplicate index values
            duplicates = df_padded.index[df_padded.index.duplicated()]
            # convert to name-value dict
            duplicate_map = {k: v for k, v in zip(duplicates.names, duplicates.values)}

            raise ValueError(f"Duplicates found: {duplicate_map}")


        # apply post script if provided
        if post_script:
            # execute post script
            self._execute_post_script(df_padded, post_script)

        return df_padded

    def _evaluate_expressions(self, df, mapping, index_columns):
        # a column header can contain a python expression. Evaluate these now that we've read all values from CSV.

        # drop index, so that we can use index columns in expressions
        if index_columns:
            df.reset_index(inplace=True)

        # store original column names so we can restore them later
        original_column_names = df.columns

        # remove foreign keys and modifiers in column names so we can use the bare names in expressions
        df.columns = df.columns.str.replace(r'\[.*\]', '', regex=True)
        df.columns = df.columns.str.replace(r'\(.*\)', '', regex=True)

        # now that we have all data, we can evaluate column expressions.
        for column in mapping['columns']:
            if 'exp' in column:
                # get column name, assuming a single attribute
                column_name = column['attributes'][0]['name']

                # evaluate expression of the form <target_column>=<expression>
                expression = f"{column_name}={column['exp']}"
                df.eval(expression, inplace=True)

        # restore column names
        df.columns = original_column_names

        # restore index
        if index_columns:
            df.set_index(index_columns, inplace=True)

        # remove skip columns, because we've evaluated expressions so we no longer need them. Match column name with 'skip=true':
        df.drop(columns=[column for column in df.columns if 'skip=true' in column], errors='ignore', inplace=True)

    def _read_from_db(self, mapping, where_clause, set_index=False):
        # get sqlalchemy engine from context
        engine = cnx_context.engine

        # get enabled and unique columns and column types
        column_names = HeaderCompiler().compile_list(mapping)
        index_columns = HeaderCompiler().compile_list_unique(mapping)
        column_types = TypesCompiler().compile(mapping, column_names)

        # read dataframe from DB
        query = self._create_select_query(mapping, where_clause)
        df = pd.read_sql_query(query, engine)

        # set headers, they must equal the request headers for comparison
        df.columns = column_names

        # set index columns, if header contains unique columns
        if set_index and index_columns:
            df.set_index(index_columns, inplace=True)

        # apply converters after reading from DB, because read_sql_query() doesn't support converters
        converters = column_types['read_db_converters']
        self._apply_converters(column_names, converters, df)

        # force pandas to not convert int columns to float if they contain NaNs
        df = df.convert_dtypes()

        return df

    def _apply_converters(self, column_names, converters, df):
        # iterate the columns to apply converters one by one to either value or index column
        for column in column_names:
            if column in converters:
                if column in df.columns:

                    # apply this converter to value column
                    df[column] = df[column].apply(converters[column])
                elif column in df.index.names:

                    # get the index column
                    index_column = df.index.levels[df.index.names.index(column)]

                    # apply this converter to the index column
                    df.index = df.index.set_levels(index_column.map(converters[column]), level=column)

    def _compare(self, df_request, df_db, insert, update, delete):
        # remove columns with empty names. Don't do this when reading from DB, because in get_table request we also want empty columns
        df_db = df_db.drop(columns=[''], errors='ignore')

        # find rows to insert
        inserted_indices = set(df_request.index.values).difference(df_db.index.values)
        inserts = df_request[df_request.index.isin(inserted_indices)]
        inserts.reset_index(inplace=True)

        # move __line__ to become the left most column after index reset
        inserts = self._move_line_to_front(inserts)

        # find rows to delete
        deleted_indices = set(df_db.index.values).difference(df_request.index.values)
        deletes = df_db[df_db.index.isin(deleted_indices)]
        deletes.reset_index(inplace=True)

        # find rows to update. Left is from request, right is from database
        left = df_request[~df_request.index.isin(inserted_indices)]
        right = df_db[~df_db.index.isin(deleted_indices)]

        # Sort the right DataFrame to match the index of the left DataFrame
        right_df_sorted = right.reindex(left.index)

        # left (from request) has __line__ column. Create a copy without line numbers for comparison
        left_no_line = left.drop(columns=['__line__'])

        # get mask of cells that are N/A or '' in left and right
        mask = (left_no_line.isna() | left_no_line.eq('')) & (right_df_sorted.isna() | right_df_sorted.eq(''))

        # replace N/A or '' with None using mask to ignore differences in N/A or ''
        left_no_line = left_no_line.mask(mask, None)
        right_df_sorted = right_df_sorted.mask(mask, None)

        # compare remaining rows
        updates = left_no_line.compare(right_df_sorted)

        # insert __line__ column, recovering line numbers from left by index
        updates['__line__'] = left.loc[updates.index]['__line__']

        # reset index for updates to match the insert and deletes
        updates.reset_index(inplace=True)

        # move __line__ to become the left most column
        updates = self._move_line_to_front(updates)

        # return result
        return inserts if insert else DataFrame(), updates if update else DataFrame(), deletes if delete else DataFrame()

    def _move_line_to_front(self, df):
        # move __line__ to become the left most column. This has no real purpose, but it makes the dataframes more readable
         # this must also work with the multi-index dataframes coming from the compare function
        # get __line__ column as a series
        line = df['__line__']
        # drop __line__ column, use level parameter to improve performance
        df = df.drop(columns=['__line__'])
        # insert __line__ column as first column
        df.insert(0, '__line__', line)
        return df


    def get_select_statement(self, table_name, header, where_clause=None):
        # get cnx from context
        cnx = cnx_context.cnx

        # parse header to build mapping
        mapping = HeaderParser(get_metadata(cnx), table_name).parse_csv(header)

        # Read data from the database table using pandas
        query = self._create_select_query(mapping, where_clause)

        return str(query)

    def _create_select_query(self, mapping, where_clause):
        # translate syntax tree to select query
        return SelectCompiler().compile(mapping, where_clause)

    def get_header_json(self, table_name, header=None):
        # get mapping, merged if needed
        mapping = self._get_header_mapping(table_name, header)

        # compile into header
        json_header = HeaderCompiler().compile_json(mapping)

        return json_header

    def get_header_csv(self, table_name, header=None):
        # get mapping, merged if needed
        mapping = self._get_header_mapping(table_name, header)

        # compile into header
        csv_header = HeaderCompiler().compile_csv(mapping)

        return csv_header

    def _get_header_mapping(self, table_name, header=None):
        # get a MetaData instance from context
        metadata = get_metadata(cnx_context.cnx)

        # create default mapping from table
        mapping = OdooHeaderParser(metadata, cnx_context.cr).parse(table_name)

        if not header:
            # return full mapping
            return mapping

        # parse header to build syntax tree
        parsed_header = HeaderParser(metadata, table_name).parse_csv(header)

        # merge headers
        merged_mapping = HeaderMerger().merge(mapping, parsed_header)

        return merged_mapping

    def _execute_sql(self, query_executors, commit=False):
        # get cursor from context
        cr = cnx_context.cr

        # execute queries, rerun until exhausted
        result = self._eat_sleep_repeat(query_executors, cr)

        # commit if requested
        if commit:
            cnx_context.cnx.commit()

            # registry may not be availabe during unit tests
            if hasattr(cnx_context, 'registry'):
                # invalidate caches to avoid stale values coming from cache
                cnx_context.registry.clear_cache()

        return result

    def _eat_sleep_repeat(self, query_executors, cr):
        # execute in rounds until no new successful queries are found

        # create result lists
        completed = []
        failed = []

        # copy query executors list
        remaining = query_executors.copy()

        done = False

        while not done:
            new_completed_executors = []
            new_completed_results = []
            # iterate query executors
            for query_executor in remaining:
                # create or replace savepoint
                self.create_savepoint()
                # delegate execution to query executor
                execution_result = query_executor.execute(cr)
                # if successful
                if execution_result.success:
                    # append result to list
                    new_completed_results.append(execution_result)
                    # remove from remaining
                    new_completed_executors.append(query_executor)
                else:
                    # rollback to savepoint
                    self.rollback_to_savepoint()
                    # append to failed list
                    failed.append(execution_result)
            if new_completed_results:
                # append new completed to completed list
                completed.extend(new_completed_results)
                # remove completed executors from remaining
                remaining = [qe for qe in remaining if qe not in new_completed_executors]
                # reset failed list and start again
                failed = []
            else:
                # nothing new completed, we're done
                done = True
        # return completed and failed lists
        return completed + failed


    def set_context(self, url, password):
        # create psycopg2 connection
        cnx = psycopg2.connect(url, password=password)
        # set connection and cursor in context
        cnx_context.cnx = cnx
        cnx_context.cr = cnx.cursor()

    def create_savepoint(self):
        # create savepoint
        cnx_context.cr.execute("SAVEPOINT stimula_savepoint")

    def rollback_to_savepoint(self):
        # rollback to savepoint
        cnx_context.cr.execute("ROLLBACK TO SAVEPOINT stimula_savepoint")

    def _replace_empty_columns_with_skip(self, column_names):
        # replace empty column names with skip, skip1, skip2. This is because pandas doesn't like empty column names
        i = 0
        for name in column_names:
            if name == '':
                name = 'skip' + ('' if i == 0 else str(i))
                i += 1
            yield name

    def _find_duplicate_names(self, names):
        # find duplicate names
        seen = set()
        duplicates = set()
        for name in names:
            if name in seen:
                duplicates.add(name)
            seen.add(name)
        return duplicates

    def _get_selected_columns(self, table_name, all_columns, header):
        # get cnx from context
        cnx = cnx_context.cnx

        # parse header to build syntax tree
        mapping = HeaderParser(get_metadata(cnx), table_name).parse_csv(header)

        # reconstruct original header from tree
        column_keys = HeaderCompiler().compile_list(mapping)

        # create result list for selected and unselected columns
        result_columns = []

        # initiate list for unselected columns with all default columns
        unselected_columns = all_columns.copy()

        # iterate header columns to retain order of selected items
        for key in column_keys:

            # find column with matching key
            column = next((c for c in unselected_columns if c['key'] == key), None)

            # raise error if if column is not found,
            if column is None:
                raise ValueError(f'Column {key} not found in table {table_name}')

            # set selected is true
            column['selected'] = True

            # remove from unselected columns and add to selected columns
            unselected_columns.remove(column)
            result_columns.append(column)

        # add remaining unselected columns to selected columns
        result_columns.extend(unselected_columns)

        return result_columns

    def _count_body_columns(self, body):
        # skipinitialspace must be true, otherwise it may split on comma's in strings
        df_initial = pd.read_csv(StringIO(body), skipinitialspace=True, header=None)

        # return the number of columns in the body
        return len(df_initial.columns)

    def _pad_dataframe_with_empty_columns(self, df, column_names, index_columns, converters):
        # get current column names
        current_columns = df.columns.tolist()

        # get number of columns to pad, take columns and indices into account
        pad_count = len(column_names) - df.shape[1] - df.index.nlevels

        # if there are too few columns
        if pad_count <= 0:
            # nothing to pad
            return df

        # get columns to pad
        pad_columns = column_names[-pad_count:]

        # pad the DataFrame with empty columns
        _logger.info(f"Padding DataFrame with {pad_count} empty columns")
        current_columns += pad_columns

        # be careful not to set index columns, the columns attribute must only be set to non-index column names
        df_padded = df.reindex(columns=current_columns, fill_value=None)

        # apply converters to padded columns
        for column in pad_columns:
            if column in converters:
                df_padded[column] = df_padded[column].apply(converters[column])

        # set additional index columns
        for column in pad_columns:
            if column in index_columns:
                df_padded.set_index(column, append=True, inplace=True)

        return df_padded

    def _deduplicate(self, df, index_columns):
        # reset index
        df_reset = df.reset_index()

        # Remove duplicate rows based on the index column
        df_unique = df_reset.drop_duplicates(subset=index_columns)

        # Set the column back as the index
        df_final = df_unique.set_index(index_columns)

        return df_final

    def _execute_post_script(self, df, post_script):
        # skip if no post script provided
        if post_script is None:
            return df

        # assert that post_script file exists
        assert os.path.exists(post_script), f"Post script file {post_script} not found"

        # import the post script module
        module_name = 'post_script'
        spec = importlib.util.spec_from_file_location(module_name, post_script)
        module = importlib.util.module_from_spec(spec)
        sys.modules[module_name] = module
        spec.loader.exec_module(module)

        # verify that the module has an execute function
        assert hasattr(module, 'execute'), f"Post script module {post_script} must have an execute function"

        # execute the post script
        return module.execute(df)



"""
This class adds alias names to tables and parameter names to terminal (not foreign key) columns.
The reason we shouldn't create parameter names on the fly when generating SQL queries, is that the order in
which the sql compilers process cells is hard to predict. For example, UpdateCompiler leaves parsing
of unique columns to the end (when generating the where clause). This makes it particularly hard to
create column headers with the correct parameter names.
This compiler generates parameter names in a single linear parsing round.

Author: Romke Jonker
Email: romke@rnadesign.net
"""
import copy


class AliasCompiler:
    def __init__(self):
        self._table_suffixes = {}
        self._parameter_suffixes = {}

    def compile(self, mapping):
        # must not re-use this instance
        assert not self._table_suffixes, 'AliasCompiler instance must not be re-used'
        # create deep copy of mapping to avoid modifying the original
        aliased_mapping = copy.deepcopy(mapping)
        table_name = aliased_mapping['table']
        self._table_suffixes[table_name] = 0
        [self._column(c) for c in aliased_mapping['columns']]
        return aliased_mapping

    def _column(self, column):
        # column may be empty
        if column:
            attributes = column['attributes']
            self._attributes(attributes)

    def _attributes(self, attributes):
        [self._attribute(attribute) for attribute in attributes]

    def _attribute(self, attribute):
        if not 'foreign-key' in attribute:
            # create unique parameter name so queries use predictable names
            attribute_name = attribute['name']
            parameter_name = self._get_unique_parameter_id(f'{attribute_name}')
            attribute['parameter'] = parameter_name
        else:
            foreign_key = attribute['foreign-key']
            table_name = foreign_key['table']
            alias = self._get_unique_table_id(table_name)
            # set alias
            foreign_key['alias'] = alias
            self._attributes(foreign_key['attributes'])

    def _get_unique_table_id(self, name):
        if name not in self._table_suffixes:
            self._table_suffixes[name] = 0
            return name
        # increase suffix to make unique
        self._table_suffixes[name] += 1
        return name + '_' + str(self._table_suffixes[name])

    def _get_unique_parameter_id(self, name):
        if name not in self._parameter_suffixes:
            self._parameter_suffixes[name] = 0
            return name
        # increase suffix to make unique
        self._parameter_suffixes[name] += 1
        return name + '_' + str(self._parameter_suffixes[name])



"""
This script provides a command-line interface (CLI) for interacting with the Stimula API.
It supports both local and remote invocations, allowing users to execute a variety of commands such as
authentication, listing tables, retrieving mappings, counting records, fetching tables, and posting data.

Author: Romke Jonker
Email: romke@rnadesign.net

Usage:
python cli.py <command> [options]

Commands:
- auth: Authenticate, retrieve a token and store it in a local file.
- list: List tables and their record counts.
- mapping: Retrieve mapping for a specified filter.
- count: Count records in a table based on specified criteria.
- get: Fetch a table with specified mapping and query parameters.
- post: Post data to a table from a specified file.

Options:
-h, --help        Print help message
-r, --remote      Remote API URL
-H, --host        Database host (default: localhost)
-P, --port        Database port (default: 5432)
-d, --database    Database name
-u, --user        Database username
-p, --password    Password
-k, --key         Secret key
-T, --token       Authentication token
-t, --table       Table name or filter
-q, --query       Query clause
-m, --mapping     Mapping header
-f, --file        Path to the file to post (required for post command)
-s, --skip        Number of rows to skip (default: 1)
-e, --enable      Enable flags (I, U, D, E, C)
-F, --format      Response format (choices: diff, sql)
-D, --deduplicate Deduplicate records based on primary key
-v, --version     Print version information
-V, --verbose     Increase output verbosity
-M, --transpose   Transpose the mapping
-x, --execute     Script to execute on post
"""

import argparse
import os
import sys
from importlib.metadata import version
from io import StringIO

import pandas as pd

from stimula.cli import local, remote


def main():
    parser = argparse.ArgumentParser(description='stimula - The STML CLI')
    parser.add_argument('command', help='Command to execute', choices=['auth', 'list', 'mapping', 'count', 'get', 'post', 'transpose'])
    parser.add_argument('-r', '--remote', help='Remote API URL')
    parser.add_argument('-H', '--host', help='Database host', default='localhost')
    parser.add_argument('-P', '--port', help='Database port', type=int, default=5432)
    parser.add_argument('-d', '--database', help='Database name')
    parser.add_argument('-u', '--user', help='Database username')
    parser.add_argument('-p', '--password', help='Password')
    parser.add_argument('-k', '--key', help='Secret key')
    parser.add_argument('-T', '--token', help='Authentication token')
    parser.add_argument('-t', '--table', nargs='+', help='Table name or filter')
    parser.add_argument('-q', '--query', help='Query clause')
    parser.add_argument('-m', '--mapping', help='Mapping header')
    parser.add_argument('-f', '--file', nargs='+', help='Path to the file to post', type=argparse.FileType('rb'))
    parser.add_argument('-s', '--skip', help='Number of rows to skip', type=int, default=1)
    parser.add_argument('-e', '--enable', help='Enable flags', type=validate_flags)
    parser.add_argument('-F', '--format', help='Response format', choices=['diff', 'sql', 'full'], default='full')
    parser.add_argument('-D', '--deduplicate', action='store_true', help='Deduplicate by unique key')
    parser.add_argument('-v', '--version', action='version', version=version('stimula'))
    parser.add_argument('-V', '--verbose', action='store_true', help='Increase output verbosity')
    parser.add_argument('-M', '--transpose', action='store_true', help='Transpose the mapping')
    parser.add_argument('-x', '--execute', help='Script to execute on post')
    parser.add_argument('-c', '--context', nargs='+', help='Free text to match the query results')


    args = parser.parse_args()

    try:
        execute_command(args)
    except Exception as e:
        if args.verbose:
            # print message with stack trace
            raise e
        else:
            # print message without stack trace to stderr
            print(f'Error: {e}', file=sys.stderr)

        sys.exit(1)


def execute_command(args):

    if args.command == 'transpose':
        # transpose stdin to stdout and exit
        _transpose_stdin_stdout()
        return

    if args.remote:
        # if remote is specified, use remote invoker
        invoker = remote.Invoker(args.remote)
    else:
        # otherwise, read key from environment if not provided as argument
        if not args.key:
            args.key = os.getenv('STIMULA_KEY')

        # if key is still not provided, raise an error
        assert args.key, 'Secret key must be provided for local connection. Either connect to remote API, or set --key argument, or STIMULA_KEY environment variable'

        # use local invoker
        invoker = local.Invoker(args.key, args.host, args.port)

    # try to read token from local file
    _read_token_from_file(args)

    # if auth request or no token provide
    if args.command == 'auth' or not args.token:
        # authenticate and set token
        args.token = invoker.auth(args.database, args.user, args.password)
        #     # write token to local file
        _write_token_to_file(args.token)

    # validate token and set connection context
    invoker.set_context(args.token)

    # check if mapping is provided as file name, then use the first line
    _read_mapping_from_file(args)

    if args.transpose:
        # transpose mapping
        _transpose_mapping(args)

    # use only first line of mapping
    if args.mapping:
        args.mapping = args.mapping.splitlines()[0]

    # execute command
    if args.command == 'auth':
        print(f'Token: {args.token}')
    elif args.command == 'list':
        filter = args.table[0] if args.table else None
        tables = invoker.list(filter)
        # print name and count of tables
        for table in tables:
            print(f'{table["name"]}: {table["count"]}')
    elif args.command == 'mapping':
        assert args.table, 'Table name must be provided using -t or --table flag.'
        mapping = invoker.mapping(args.table[0])
        print(mapping)
    elif args.command == 'count':
        assert args.table, 'Table name must be provided using -t or --table flag.'
        count = invoker.count(args.table[0], args.mapping, args.query)
        print(count)
    elif args.command == 'get':
        assert args.table, 'Table name must be provided using -t or --table flag.'
        table = invoker.get_table(args.table[0], args.mapping, args.query)
        print(table)
    elif args.command == 'post':
        assert args.table, 'Table name must be provided using -t or --table flag.'
        assert args.enable, 'At least one of the flags I, U, D, E, or C must be enabled. Otherwise, there\'s nothing to do.'

        # raise error if no contents are provided
        assert (args.file and len(args.file) > 0) or not sys.stdin.isatty(), 'No contents provided, either use --file or -f flag, or pipe data to stdin.'

        if args.mapping is None or args.mapping == '':
            assert args.skip > 0, 'No mapping provided and skip is zero. Specify a mapping using the --mapping or -m flag, or provide a file with a header row and --skip > 0.'
            assert not args.transpose, 'Cannot transpose mapping when reading header from data file.'
            # leave it to the server to use first line of contents as mapping

        # if verbose, print mapping
        if args.verbose:
            print(f'Mapping: {args.mapping}')

        csv = invoker.post_table(args.table, args.mapping, args.query, args.file,
                                 skiprows=args.skip,
                                 insert='I' in args.enable,
                                 update='U' in args.enable,
                                 delete='D' in args.enable,
                                 execute='E' in args.enable,
                                 commit='C' in args.enable,
                                 format=args.format,
                                 deduplicate=args.deduplicate,
                                 post_script=args.execute,
                                 context=args.context)
        print(csv)


def validate_flags(value):
    valid_letters = set("IUDEC")
    input_set = set(value)

    if not input_set.issubset(valid_letters):
        raise argparse.ArgumentTypeError(f"Invalid combination: {value}. Only the letters I, U, D, E, and C are allowed.")

    return value


def _transpose_stdin_stdout():
    # validate we have stdin
    assert not sys.stdin.isatty(), 'No input provided, use piping to provide input.'
    # read dataframe from stdin
    df = pd.read_csv(sys.stdin, header=None)
    # transpose the dataframe and write to stdout
    df.T.to_csv(sys.stdout, header=False, index=False)


def _read_token_from_file(args):
    # try to read token from local file if not provided in arguments
    if not args.token:
        try:
            with open('.stimula_token', 'r') as file:
                args.token = file.read()
        except FileNotFoundError:
            pass


def _write_token_to_file(token):
    # write token to local file
    with open('.stimula_token', 'w') as file:
        file.write(token)


def _read_mapping_from_file(args):
    # check if args.mapping is an existing file name
    if args.mapping and os.path.exists(args.mapping):
        with open(args.mapping, 'r') as file:
            # read the whole file, it may be a transposed mapping
            args.mapping = file.read()


def _transpose_mapping(args):
    assert args.mapping, 'Mapping must be provided when transposing.'
    # get string from args.mapping, remove lines starting with #
    mapping = '\n'.join(line for line in args.mapping.splitlines() if not line.startswith('#'))
    # use pandas to read the mapping string into a dataframe
    df = pd.read_csv(StringIO(mapping), header=None)
    # transpose the dataframe, dispose all but the first row and convert to csv
    args.mapping = df.T.head(1).to_csv(header=False, index=False)


if __name__ == '__main__':
    main()



"""
This is an abstract base class for authentication. It provides methods to authenticate and validate tokens.

Author: Romke Jonker
Email: romke@rnadesign.net
"""
import base64
import os

import jwt
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes, padding
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC
from abc import ABC, abstractmethod


class Auth(ABC):
    # set the secret key during instantiation
    def __init__(self, secret_key):
        # secret_key must be set
        assert secret_key, 'Secret key must be set'
        self._secret_key = secret_key

    def authenticate(self, database, username, password):
        # validate submitted credentials and obtain user ID to store in token
        uid = self._validate_submitted_credentials(database, username, password)

        # encrypt password
        encrypted_password, salt = self.encrypt(self._secret_key, password)

        payload = {"database": database, "uid": uid, "password": encrypted_password, "salt": salt}
        token = jwt.encode(payload, self._secret_key, algorithm='HS256')
        return token

    def validate_token(self, token):
        # decode token
        payload = jwt.decode(token, self._secret_key, algorithms=['HS256'])

        # get connection parameters from payload
        database, uid, encrypted_password, salt = payload['database'], payload['uid'], payload['password'], payload['salt']

        # decrypt password
        password = self.decrypt(self._secret_key, encrypted_password, salt)

        # validate token credentials and return resulting objects
        result = self._validate_token_credentials(database, uid, password)

        # return validation result
        return result

    @abstractmethod
    def _validate_submitted_credentials(self, database, username, password):
        pass

    @abstractmethod
    def _validate_token_credentials(self, database, username, password):
        pass

    def encrypt(self, secret, plaintext):
        salt = os.urandom(16)  # Generate a random salt
        dek = self._derive_key_from_secret(secret, salt)

        encrypted_secret = self._encrypt_with_derived_key(plaintext, dek)
        return base64.urlsafe_b64encode(encrypted_secret).decode(), base64.urlsafe_b64encode(salt).decode()

    def _encrypt_with_derived_key(self, plaintext, dek):
        cipher = Cipher(algorithms.AES(dek), modes.ECB(), backend=default_backend())
        encryptor = cipher.encryptor()

        padder = padding.PKCS7(algorithms.AES.block_size).padder()
        padded_data = padder.update(plaintext.encode()) + padder.finalize()

        ciphertext = encryptor.update(padded_data) + encryptor.finalize()
        return ciphertext

    def decrypt(self, secret, ciphertext, salt):
        dek = self._derive_key_from_secret(secret, base64.urlsafe_b64decode(salt))
        encrypted_data = base64.urlsafe_b64decode(ciphertext)

        decrypted_data = self._decrypt_with_derived_key(encrypted_data, dek)

        return decrypted_data

    def _decrypt_with_derived_key(self, ciphertext, dek):
        cipher = Cipher(algorithms.AES(dek), modes.ECB(), backend=default_backend())
        decryptor = cipher.decryptor()
        decrypted_data = decryptor.update(ciphertext) + decryptor.finalize()

        # Use PKCS7 unpadding
        unpadder = padding.PKCS7(algorithms.AES.block_size).unpadder()
        plaintext = unpadder.update(decrypted_data) + unpadder.finalize()

        return plaintext.decode()

    def _derive_key_from_secret(self, passphrase, salt):
        kdf = PBKDF2HMAC(
            algorithm=hashes.SHA256(),
            length=32,  # Length of the derived key
            salt=salt,
            iterations=100000,  # Adjust as needed
            backend=default_backend()
        )
        # return base64.urlsafe_b64encode(kdf.derive(passphrase.encode())).decode()
        return kdf.derive(passphrase.encode())



"""
This class parses a text format mapping header. It connects to the database to verify the existence of tables and columns
and it resolves foreign key references.

This parser is not thread safe, because it uses a stack to keep track of the current table and column. Make sure that
each thread has its own instance of this parser.

Author: Romke Jonker
Email: romke@rnadesign.net

The grammar is as follows:

cell         : columns modifiersets
             | columns
             | empty

columns      : columns : column
             | column

column       : ID ( columns )
             | ID

modifiersets : modifiersets : modifierset
             | modifierset

modifierset  : [ modifiers ]

modifiers    : modifiers : modifier
             | modifier

modifier     : ID = ID
             | ID = VALUE

"""
import csv
import io

from .header_lexer import HeaderLexer
from .sly import Parser


class HeaderParser(Parser):
    # CSV headers that are treated as boolean
    BOOLEAN_HEADERS = ['unique', 'skip']
    _lexer = HeaderLexer()

    def __init__(self, metadata, table_name):
        self.metadata = metadata
        self._table_name = table_name

    # Get the token list from the lexer (required)
    tokens = HeaderLexer.tokens

    # main parse function that starts with decoding CSV style headers
    def parse_csv(self, header):
        # resolve table
        table = self._resolve_table(self._table_name)

        # an empty string can either be an empty header or a header with a single empty column. Let's return an empty header in that case.
        if header.strip() == '':
            return {'table': self._table_name}

        # decode CSV style header
        csv_file = io.StringIO(header)

        # Create a CSV reader object
        csv_reader = csv.reader(csv_file, delimiter=',', quotechar='"', skipinitialspace=True)

        # Parse the single line into a list of cells
        cells = next(csv_reader)

        # push table name on stack
        self.table_stack = [(table, None)]

        # parse the cells
        parsed_cells = [self.parse(HeaderParser._lexer.tokenize(cell)) for cell in cells]

        # pop the stack and verify it's empty
        table, _ = self.table_stack.pop()
        assert not self.table_stack, 'table stack should be empty after parsing'

        # for extension relations, we still have to resolve the foreign key table using the modifiers
        for cell in parsed_cells:
            self._complete_extension_foreign_key(cell, cell)

        # Verify that all attributes exist in the table by checking that they have a type.
        # For columns marked as 'skip=true', the column does not have to exist in the table.
        # We can't do this check during parsing of the attribute, bec/ then the modifiers aren't known yet.
        for cell in parsed_cells:
            if not cell.get('skip'):
                for a in cell.get('attributes', []):
                    self._verify_attribute(a, table.name)

        return {'table': table.name, 'columns': parsed_cells}

    def _verify_attribute(self, attribute, table):
        if 'foreign-key' in attribute:
            foreign_key = attribute['foreign-key']
            for a in foreign_key.get('attributes', []):
                self._verify_attribute(a, foreign_key['table'])
        else:
            if 'type' not in attribute:
                raise ValueError(f"Column '{attribute['name']}' not found in table '{table}'")

    def _complete_extension_foreign_key(self, modifiers, cell_or_foreign_key):
        # for extension relations, we need to complete the foreign key once we have parsed the modifiers
        for attribute in cell_or_foreign_key.get('attributes', []):
            if 'foreign-key' in attribute:
                foreign_key = attribute['foreign-key']
                if 'table' in foreign_key:
                    # this is a proper foreign key, recurse to see if it has extension relations
                    self._complete_extension_foreign_key(modifiers, foreign_key)

                else:
                    # this may be an extension relation, we need to resolve the foreign key table and column

                    # find foreign table name in cell and remove the attribute
                    if 'table' not in cell_or_foreign_key:
                        raise ValueError(f"Column '{attribute['name']}' is not a foreign key and no 'table' specified in modifiers")
                    table = self._resolve_table(modifiers['table'])
                    del modifiers['table']

                    # find referred column in cell and remove the attribute
                    if 'name' not in modifiers:
                        raise ValueError(f"Column '{attribute['name']}' is not a foreign key and no referred 'name' specified in modifiers")
                    column_name = modifiers['name']
                    del modifiers['name']

                    if column_name not in table.columns:
                        raise ValueError(f"Column '{column_name}' not found in table '{table}'")
                    foreign_key['table'] = table.name
                    foreign_key['name'] = column_name

                    # Odoo's ir_model_data requires a qualifier name. Find it in cell and remove the attribute.
                    if 'qualifier' not in modifiers:
                        raise ValueError(f"Column '{attribute['name']}' is not a foreign key and no 'qualifier' specified in modifiers")
                    foreign_key['qualifier'] = modifiers['qualifier']
                    del modifiers['qualifier']

                    # mark the foreign key as extension
                    foreign_key['extension'] = True

                    # also fix attribute types
                    for a in foreign_key.get('attributes', []):
                        if a['name'] not in table.columns:
                            raise ValueError(f"Column '{a['name']}' not found in table '{table}'")
                        column_type = str(table.columns[a['name']].type).lower()
                        a['type'] = column_type

    @_('columns modifiersets')
    def cell(self, p):
        return {'attributes': p.columns, 'enabled': True, **p.modifiersets}

    @_('columns')
    def cell(self, p):
        return {'attributes': p.columns, 'enabled': True}

    @_('empty')
    def cell(self, p):
        return {}

    @_('columns COLON column')
    def columns(self, p):
        return p.columns + [p.column]

    @_('column')
    def columns(self, p):
        return [p.column]

    @_('ID push LPAREN columns RPAREN')
    def column(self, p):
        table, column = self.table_stack.pop()

        if table is None:
            #     this may happen for extension relations, we need to resolve foreign table and column we have parsed the modifiers
            return {'name': p.ID, 'foreign-key': {'attributes': p.columns}}

        return {'name': p.ID, 'foreign-key': {'table': table.name, 'name': column, 'attributes': p.columns}}

    @_('')
    def push(self, p):
        # push target table name on stack
        column = p[-1]
        target = self._resolve_foreign_key_table(column)
        self.table_stack.append(target)

    @_('ID')
    def column(self, p):
        # verify column exists
        table, _ = self.table_stack[-1]
        if table is None:
            # this happens for extension relations, we don't know the table until we've read the modifiers
            return {'name': p.ID}

        if p.ID not in table.columns:
            # column does not exist, this is fine for [skip=true] columns
            return {'name': p.ID}

        column = table.columns[p.ID]

        return {
            'name': str(column.key),
            'type': str(column.type).lower()
        }

    @_('modifiersets modifierset')
    def modifiersets(self, p):
        return p.modifiersets | p.modifierset

    @_('modifierset')
    def modifiersets(self, p):
        return p.modifierset

    @_('LBRACK modifiers RBRACK')
    def modifierset(self, p):
        return p.modifiers

    @_('modifiers COLON modifier')
    def modifiers(self, p):
        return p.modifiers | p.modifier

    @_('modifier')
    def modifiers(self, p):
        return p.modifier

    @_('ID EQUALS ID')
    @_('ID EQUALS VALUE')
    def modifier(self, p):
        key = p[0]
        if key in self.BOOLEAN_HEADERS:
            # boolean modifier
            value = str(p[2]).lower() == 'true'
        else:
            value = p[2]

        return {key: value}

    @_('ID EQUALS QUOTED_VALUE')
    def modifier(self, p):
        key = p[0]
        # strip quotes
        value = p[2][1:-1]

        return {key: value}

    @_('')
    def empty(self, p):
        pass

    def _resolve_table(self, table_name):
        table = self.metadata.tables.get(table_name)
        if table is None:
            raise ValueError(f"Table '{table_name}' not found")
        return table

    def error(self, token):
        msg = 'Parse error'
        if token:
            # if token has value attribute
            value = getattr(token, 'value', None)
            if value:
                msg += f': encountered \'{value}\''

            # if token has index attribute
            index = getattr(token, 'index', 0)
            if index:
                msg += f' at index {index}'

        # fail fast
        raise ValueError(msg)

    def _resolve_foreign_key_table(self, column_name):
        # get referred column
        table, _ = self.table_stack[-1]
        if column_name not in table.columns:
            raise ValueError(f"Column '{column_name}' not found in table '{table.name}'")
        column = table.columns[column_name]
        # only know how to deal with a single foreign key per column
        if len(column.foreign_keys) != 1:
            # this is either an error condition or an extension relation, with the foreign key in the extension table. We can tell once we have parsed the modifiers.
            return None, column_name

        foreign_key = list(column.foreign_keys)[0]
        table = foreign_key.column.table
        column = foreign_key.column.name
        return table, column



"""
This class compiles a mapping into an update query.

Author: Romke Jonker
Email: romke@rnadesign.net
"""
from .alias_compiler import AliasCompiler
from .insert_compiler import ForeignWhereClauseCompiler, FromClauseCompiler


class UpdateCompiler:
    """
        update c
        set c2 = c0.c0, c3 = b.b0, c4 = b0.b0
        from c as c0, b, b as b0
        left join a on b0.b2 = a.a0
        where c.c1 = 'c4' and c0.c1 = 'c2' and b.b1 = 'b1' and a.a1 = 'a2';
    """

    def compile(self, mapping):
        aliased_mapping = AliasCompiler().compile(mapping)
        update_clause = UpdateClauseCompiler().compile(aliased_mapping)
        from_clause = FromClauseCompiler(False).compile(aliased_mapping)
        where_clause = WhereClauseCompiler().compile(aliased_mapping)
        foreign_where_clause = ForeignWhereClauseCompiler(False).compile(aliased_mapping)

        result = f'{update_clause}{from_clause}{where_clause}'
        if foreign_where_clause:
            result += ' and ' + foreign_where_clause
        return result


class UpdateClauseCompiler:
    def compile(self, mapping):
        table = mapping['table']

        # compile and filter
        clause_lists = [self._column(c, table) for c in mapping['columns']]
        clauses = [clause for clause_list in clause_lists for clause in clause_list]

        # comma separate cells
        return f'update {table} set ' + ', '.join(clauses)

    def _column(self, column, table):
        unique = column.get('unique', False)
        # can't update unique columns
        if unique:
            return []

        return [self._attribute(attribute) for attribute in column['attributes']]

    def _attribute(self, attribute):
        # if no foreign key, get value from parameter
        if not 'foreign-key' in attribute:
            column_name = attribute['name']
            parameter_name = attribute['parameter']
            return f'{column_name} = :{parameter_name}'
        else:
            foreign_key = attribute['foreign-key']
            column_name = attribute['name']
            # but table names may need an alias
            target_table = foreign_key['table']
            target_alias = foreign_key.get('alias', target_table)
            target_name = foreign_key['name']
            # no need to recurse
            return f'{column_name} = {target_alias}.{target_name}'


class WhereClauseCompiler:
    def compile(self, mapping):
        table_name = mapping['table']
        # filter columns
        clause_lists = [self._column(c, table_name) for c in mapping['columns']]
        clauses = [clause for clause_list in clause_lists for clause in clause_list]

        # must not be empty
        if not clauses:
            raise Exception('Header must have at least one unique column')

        return ' where ' + ' and '.join(clauses)

    def _column(self, column, table_name):
        unique = column.get('unique', False)
        if not unique:
            return []

        return [self._attribute(attribute, table_name) for attribute in column['attributes']]

    def _attribute(self, attribute, alias):
        column_name = attribute['name']
        # this compiler is only for the base table, don't recurse foreign keys

        if 'foreign-key' in attribute:
            # restrict the foreign key to the base table
            foreign_key = attribute['foreign-key']
            return f'{alias}.{column_name} = {foreign_key["alias"]}.{foreign_key["name"]}'

        # if no foreign key, restrict base table to value from parameter
        parameter_name = attribute.get('parameter', f'{column_name}')
        return f'{alias}.{column_name} = :{parameter_name}'



"""
This class compiles a mapping into a select query.

Author: Romke Jonker
Email: romke@rnadesign.net
"""
import re

from .alias_compiler import AliasCompiler


class SelectCompiler:
    """
        select c.c1, c0.c1, b.b1, a.a1
        from c
        left join c as c0 on c.c2 = c0.c0
        left join b on c.c3 = b.b0
        left join b as b0 on c.c4 = b0.b0
        left join a on b0.b2=a.a0
        order by c.c1
    """

    def compile(self, mapping, where_clause=None):
        """
        Compiles a mapping into a select query
        :param mapping: the mapping
        :param where_clause: a free where clause for the caller to specify
        :return: the select query
        """
        table = mapping['table']
        aliased_mapping = AliasCompiler().compile(mapping)
        select_clause = SelectClauseCompiler().compile(aliased_mapping)
        join_clause = JoinClauseCompiler().compile(aliased_mapping)
        order_by_clause = OrderByClauseCompiler().compile(aliased_mapping)

        # compile [filter="...$..."] headers into a where clause, replacing '$' with the column name
        where = WhereClauseCompiler().compile(aliased_mapping, where_clause)
        return f'{select_clause} from {table}{join_clause}{where}{order_by_clause}'

    def compile_count_query(self, mapping, where_clause=None):
        table = mapping['table']
        aliased_mapping = AliasCompiler().compile(mapping)
        select_clause = 'select count(*)'
        join_clause = JoinClauseCompiler().compile(aliased_mapping)
        where = f' where {where_clause}' if where_clause else ''
        return f'{select_clause} from {table}{join_clause}{where}'

    def get_model_name(self, alias):
        # this method gets the model name to use in the extension clause
        # current solution is based on the alias, which is not very robust
        # alias is of the form 'xxx_xxx_xxx_d', we need to convert that into 'xxx.xxx.xxx'

        # remove '_0' postfix from alias using regex if it exists
        alias = re.sub(r'_\d+$', '', alias)
        return alias.replace('_', '.')



class SelectClauseCompiler:
    def compile(self, mapping):
        table_name = mapping['table']
        # comma separate cells. Skip cells with skip=true modifier. We need those when reading CSV, but not when reading from DB
        return 'select ' + ', '.join([self._column(c, table_name) for c in mapping['columns'] if not c.get('skip')])

    def _column(self, column, table_name):
        # column may be empty
        if not column:
            return '\'\''

        attributes = column['attributes']

        # empty column, select empty string
        if not attributes:
            return '\'\''

        # get the attributes for this column
        selects = self._attributes(attributes, table_name)

        # if there's more than one attribute, we must cast jsonb columns to string
        cast_selects = self._cast_selects_if_needed(selects)

        # colon separate attributes
        return ' || \':\' || '.join(cast_selects)

    def _attributes(self, attributes, alias):
        attribute_lists = [self._attribute(attribute, alias) for attribute in attributes]
        # flatten list of list into a list
        return [attribute for attribute_list in attribute_lists for attribute in attribute_list]

    def _attribute(self, attribute, alias):
        # if this is not a foreign key, return alias and column name
        if not 'foreign-key' in attribute:
            column_name = attribute['name']
            # also return the source attribute, so we can check the type later
            return [(f'{alias}.{column_name}', attribute)]
        else:
            foreign_key = attribute['foreign-key']
            # but table names may need an alias
            attributes = foreign_key['attributes']
            target_name = foreign_key['table']
            target_alias = foreign_key.get('alias', target_name)
            # recurse into foreign key table to get the attributes
            return self._attributes(attributes, target_alias)

    def _cast_selects_if_needed(self, selects):
        # postgres can't concatenate jsonb columns to anything, so cast to string
        cast_selects = []
        for select in selects:
            # select is a tuple of (column, attribute)
            if len(selects) > 1 and select[1].get('type') == 'jsonb':
                cast_selects.append(f'{select[0]}::text')
            else:
                cast_selects.append(select[0])
        return cast_selects


class JoinClauseCompiler:
    def compile(self, mapping):
        table_name = mapping['table']
        # glue cells together
        return ''.join([self._column(c, table_name) for c in mapping['columns']])

    def _column(self, column, table_name):
        # column may be empty
        if not column:
            return ''
        return self._attributes(column['attributes'], table_name)

    def _attributes(self, attributes, alias):
        attribute_list = [self._attribute(attribute, alias) for attribute in attributes]
        return ''.join(attribute_list)

    def _attribute(self, attribute, alias):
        # if no foreign key, return empty string
        if not 'foreign-key' in attribute:
            return ''
        else:
            foreign_key = attribute['foreign-key']
            # but table names may need an alias
            source_name = attribute['name']
            target_name = foreign_key['name']
            target_table = foreign_key['table']
            target_alias = foreign_key.get('alias', target_table)

            # Assume we need a left join. This may require a modifier at some point.
            join_clause = f' left join {target_table}'

            # only add 'as alias' if it's different from table name
            if target_alias != target_table:
                join_clause += f' as {target_alias}'
            join_clause += f' on {alias}.{source_name} = {target_alias}.{target_name}'

            # if this is an Odoo style extension relation, then we need to filter by qualifier (module) and table (model)
            if 'qualifier' in foreign_key:
                qualifier = foreign_key['qualifier']
                # assume for now that alias is the table name. This is fine as long as we're not joining the same table multiple times
                table_name = SelectCompiler().get_model_name(alias)
                join_clause += f' and {target_alias}.model = \'{table_name}\' and {target_alias}.module = \'{qualifier}\''

            attributes = foreign_key['attributes']
            return join_clause + self._attributes(attributes, target_alias)


class OrderByClauseCompiler:
    def compile(self, mapping):
        table_name = mapping['table']
        # clauses to order by
        clause_lists = [self._column(c, table_name) for c in mapping['columns']]
        clauses = [clause for clause_list in clause_lists for clause in clause_list]

        # nothing to order by
        if not clauses:
            return ''

        return ' order by ' + ', '.join(clauses)

    def _column(self, column, table_name):
        unique = column.get('unique', False)
        if not unique:
            return []

        return self._attributes(column['attributes'], table_name)

    def _attributes(self, attributes, alias):

        attribute_list = [self._attribute(attribute, alias) for attribute in attributes]

        return attribute_list

    def _attribute(self, attribute, alias):
        # if no foreign key, get value from parameter
        if not 'foreign-key' in attribute:
            column_name = attribute['name']
            return f'{alias}.{column_name}'
        else:
            foreign_key = attribute['foreign-key']
            # but table names may need an alias
            attributes = foreign_key['attributes']
            target_name = foreign_key['table']
            target_alias = foreign_key.get('alias', target_name)
            return ', '.join(self._attributes(attributes, target_alias))

class WhereClauseCompiler:
    def compile(self, mapping, where_clause):
        """
        Compiles [select="...$..."] modifiers into a where clause
        :param mapping:
        :param where_clause: additional where clause to append
        :return: the where clause
        """
        table_name = mapping['table']

        # get conditions from columns with filter modifiers
        conditions = [self._column(c, table_name) for c in mapping['columns'] if c.get('filter')]

        # append user-defined where clause
        if where_clause:
            conditions.append(where_clause)

        if not conditions:
            return ''

        # return where clause with conditions joined by ' and '
        return ' where ' + ' and '.join(conditions)

    def _column(self, column, table_name):
        # column may be empty, have no attributes and/or have no filter
        if not column or not column.get('attributes') or not column.get('filter'):
            return []

        attributes = column['attributes']

        # get the attributes for this column
        attributes_names = self._attributes(attributes, table_name)

        # a filter column must have exactly one attribute
        assert len(attributes_names) == 1, f'Filtered column {attributes} must have exactly one attribute, got {len(attributes_names)}'

        # replace $ placeholder with attribute name
        return column['filter'].replace('$', attributes_names[0])

    def _attributes(self, attributes, alias):
        attribute_lists = [self._attribute(attribute, alias) for attribute in attributes]
        # flatten list of list into a list
        return [attribute for attribute_list in attribute_lists for attribute in attribute_list]

    def _attribute(self, attribute, alias):
        # if this is not a foreign key, return alias and column name
        if not 'foreign-key' in attribute:
            column_name = attribute['name']
            # also return the source attribute, so we can check the type later
            return [f'{alias}.{column_name}']
        else:
            foreign_key = attribute['foreign-key']
            # but table names may need an alias
            attributes = foreign_key['attributes']
            target_name = foreign_key['table']
            target_alias = foreign_key.get('alias', target_name)
            # recurse into foreign key table to get the attributes
            return self._attributes(attributes, target_alias)



"""
This script provides the context for the database connection.

Author: Romke Jonker
Email: romke@rnadesign.net
"""
import threading

from sqlalchemy import create_engine, MetaData

cnx_context = threading.local()


def get_metadata(cnx):
    engine = create_engine('postgresql://', creator=lambda: cnx)

    # Create a MetaData instance
    metadata = MetaData()

    # Reflect the existing database schema
    metadata.reflect(bind=engine)

    return metadata



"""
This class compiles a mapping into a csv-style file header format

Author: Romke Jonker
Email: romke@rnadesign.net
"""

class HeaderCompiler:
    # attributes to include as UI directives
    UI_ATTRIBUTES = ['enabled', 'primary-key', 'unique', 'in-use', 'default']
    # attributes to exclude from the modifiers list
    NON_MODIFIER_ATTRIBUTES = ['attributes', 'enabled', 'primary-key', 'foreign-key', 'in-use', 'default']

    def compile_csv(self, mapping):
        # only return enabled columns
        if 'columns' not in mapping:
            return ''
        return ', '.join([self._column(c) for c in mapping['columns'] if c.get('enabled', False)])

    def compile_csv_unique(self, mapping):
        # only return unique columns
        if 'columns' not in mapping:
            return ''
        return ', '.join([self._column(c) for c in mapping['columns'] if c.get('unique', False)])

    def compile_list(self, mapping, enabled=False, include_skip=False):
        # if enabled, only return enabled columns.
        # if include_skip, then include skip columns. This is useful when reading from CSV
        if 'columns' not in mapping:
            return []
        return [self._column(c) for c in mapping['columns'] if (not enabled or c.get('enabled', False)) and (include_skip or not c.get('skip'))]

    def compile_list_unique(self, mapping):
        if 'columns' not in mapping:
            return []
        return [self._column(c) for c in mapping['columns'] if c.get('unique', False)]

    def compile_list_non_unique(self, mapping):
        # return list of non-unique columns
        if 'columns' not in mapping:
            return []
        # remove unique columns, also remove 'skip' columns because we don't want to write them to DB
        columns = [self._column(c) for c in mapping['columns'] if not c.get('unique', False) and not c.get('skip', False)]
        # remove empty columns
        return [c for c in columns if c]

    def _column(self, column):
        # empty column does not have attributes
        if 'attributes' not in column:
            return ''
        attributes, _ = self._attributes(column['attributes'])
        modifiers = self._modifiers(column)
        return attributes + modifiers

    def _attributes(self, attributes):
        keys, types = zip(*[self._attribute(a) for a in attributes])
        type = types[0] if len(set(types)) == 1 else None
        return ':'.join(keys), type

    def _attribute(self, attribute):
        if 'foreign-key' in attribute:
            key, type = self._foreign_key(attribute['foreign-key'])
            return attribute['name'] + key, type

        type = attribute.get('type', None)
        return attribute['name'], type

    def _foreign_key(self, foreign_key):
        key, type = self._attributes(foreign_key['attributes'])
        return f"({key})", type

    def _modifiers(self, column):
        # use an exclusion list, so we can add more attributes later
        modifiers = [self._modifier(key, column[key]) for key in sorted(column.keys()) if key not in self.NON_MODIFIER_ATTRIBUTES]
        if not modifiers:
            return ''
        return '[' + ': '.join(modifiers) + ']'

    def _modifier(self, key, value):
        # if type is not yet string, then convert to string and lowercase, so we get 'true' and not 'True'
        if not isinstance(value, str):
            value = str(value).lower()

        # need to quote if value has any of the following: ' "\'='
        if any(c in value for c in [' ', '"', "'", '=']):
            return f'{key}="{value}"'
        return f'{key}={value}'

    def compile_json(self, mapping):
        result = {
            'table-name': mapping['table']
        }
        # compile columns into json
        if 'columns' in mapping:
            result['columns'] = [self._column_json(c) for c in mapping['columns']]
        return result

    def _column_json(self, column):
        key_without_modifiers, type = self._attributes(column['attributes'])
        modifiers = self._modifiers_json(column)

        # add modifiers to key
        key = key_without_modifiers + self._modifiers(column)

        if not type:
            return {'key': key, **modifiers}
        return {'key': key, 'type': type, **modifiers}

    def _modifiers_json(self, column):
        # copy column attributes that are UI directives
        modifiers = {key: column[key] for key in sorted(column.keys()) if key in self.UI_ATTRIBUTES}
        # foreign-key is a special case, check if any of the attributes have foreign keys
        if [a for a in column['attributes'] if 'foreign-key' in a]:
            modifiers['foreign-key'] = True

        return modifiers



"""
This class generates delete statements based on a mapping.

Author: Romke Jonker
Email: romke@rnadesign.net
"""
class DeleteCompiler:
    """
        delete from c
        using c
        left join b on c.c0 = b.b0
        where c.c1 = :c1 and b.b1 = :b1;
    """

    def compile(self, mapping):
        table_name = mapping['table']
        using_clause = UsingClauseCompiler().compile(mapping)
        where_clause = WhereClauseCompiler().compile(mapping)
        foreign_where_clause = ForeignWhereClauseCompiler().compile(mapping)

        result = f'delete from {table_name}{using_clause}{where_clause}'
        if foreign_where_clause:
            result += ' and ' + foreign_where_clause

        return result


class UsingClauseCompiler:
    # this class is similar to FromClauseCompiler, but it only uses unique columns
    def compile(self, mapping):
        # get table
        table_name = mapping['table']

        clause_lists = [self._column(c, table_name) for c in mapping['columns']]
        clauses = [clause for clause_list in clause_lists for clause in clause_list]

        if not clauses:
            return ''
        return f' using ' + ', '.join(clauses)

    def _column(self, column, table_name):
        # no need to list non-unique columns as 'using' table
        if not column.get('unique', False):
            return []

        return self._attributes(column['attributes'], table_name, True)

    def _attributes(self, attributes, source_alias, is_root):
        # get attributes that contains a foreign key
        foreign_key_attributes = [a for a in attributes if 'foreign-key' in a]

        return [self._attribute(attribute, source_alias, is_root) for attribute in foreign_key_attributes]

    def _attribute(self, attribute, source_alias, is_root_table):

        from_clause = ''

        # if not root, then assume we need a left join. This may require a modifier at some point.
        if not is_root_table:
            from_clause += ' left join '

        # get foreign key
        foreign_key = attribute['foreign-key']

        # add target table
        from_clause += foreign_key['table']

        # add alias
        target_name = foreign_key['table']
        target_alias = foreign_key.get('alias', target_name)
        if target_alias != target_name:
            from_clause += f' as {target_alias}'

        # if not root then add on clause
        if not is_root_table:
            from_clause += f' on {source_alias}.{attribute["name"]} = {target_alias}.{foreign_key["name"]}'

        # recurse
        return from_clause + ''.join(self._attributes(foreign_key['attributes'], target_alias, False))


class WhereClauseCompiler:
    # This compiler create the usual where clause, but adds statements to restrict the 'using' tables
    def compile(self, mapping):
        table_name = mapping['table']
        # filter columns
        clause_lists = [self._column(c, table_name) for c in mapping['columns']]
        clauses = [clause for clause_list in clause_lists for clause in clause_list]

        # must not be empty
        if not clauses:
            raise Exception('Header must have at least one unique column')

        return ' where ' + ' and '.join(clauses)

    def _column(self, column, table_name):
        unique = column.get('unique', False)
        if not unique:
            return []

        return [self._attribute(attribute, table_name) for attribute in column['attributes']]

    def _attribute(self, attribute, alias):
        # this compiler is only for the base table, don't recurse foreign keys
        if not 'foreign-key' in attribute:
            column_name = attribute['name']
            parameter_name = attribute.get('parameter', f'{column_name}')
            return f'{alias}.{column_name} = :{parameter_name}'

        foreign_key = attribute['foreign-key']
        target_name = foreign_key['table']
        target_alias = foreign_key.get('alias', target_name)
        target_column = foreign_key['name']
        return f'{alias}.{attribute["name"]} = {target_alias}.{target_column}'


class ForeignWhereClauseCompiler:

    def __init__(self):
        self._aliases = {}

    def compile(self, mapping):
        # glue cells together
        table_name = mapping['table']
        clauses = [self._column(c, table_name) for c in mapping['columns']]

        # filter and flatten
        filtered_clauses = [clause for cell in clauses for clause in cell if clause]
        if not filtered_clauses:
            return ''
        return ' and '.join(filtered_clauses)

    def _column(self, column, table):
        # no need to list non-unique columns as 'using' table
        if not column.get('unique', False):
            return []

        where_clause_columns = self._attributes(column['attributes'], table, True)
        return [c for c in where_clause_columns if c]

    def _attributes(self, attributes, source_alias, is_root):
        return [self._attribute(attribute, source_alias, is_root) for attribute in attributes]

    def _attribute(self, attribute, alias, is_root_table):
        if not 'foreign-key' in attribute:

            # only add where clauses for joined tables, but do register for alias
            if is_root_table:
                return ''
            parameter_name = attribute.get('parameter', f'{attribute["name"]}')
            return f'{alias}.{attribute["name"]} = :{parameter_name}'

        foreign_key = attribute['foreign-key']
        target_name = foreign_key['table']
        target_alias = foreign_key.get('alias', target_name)

        # recurse
        return ' and '.join(self._attributes(foreign_key['attributes'], target_alias, False))



"""
This class assists users in creating a default mapping based on an existing data model.
It assumes an Odoo-flavored data model by excluding columns that are typically used by Odoo for internal
purposes.

Author: Romke Jonker
Email: romke@rnadesign.net
"""
from sqlalchemy import UniqueConstraint, PrimaryKeyConstraint, select, func
from sqlalchemy.dialects.postgresql import JSONB

ODOO_SYSTEM_COLUMNS = ['create_date', 'create_uid', 'write_date', 'write_uid']


class OdooHeaderParser:

    # take metadata and cursor as instance variables
    def __init__(self, metadata, cursor):
        self._metadata = metadata
        self._cr = cursor
        self._aliases = {}

    def parse(self, table_name):
        # Get the table object
        table = self._metadata.tables.get(table_name)

        if table is None:
            raise ValueError('Table not found: ' + table_name)

        return {'table': table.name, 'columns': self._columns(table)}

    def _columns(self, table):
        # get in use columns once, in a single query
        in_use_columns = self._get_in_use_columns(table)

        # get all columns
        all_columns = [self._column(c, in_use_columns) for c in table.columns]

        # We must have at least one column enabled. If no columns are enabled, then enable all columns.
        if not any([c.get('enabled', False) for c in all_columns]):
            for c in all_columns:
                c['enabled'] = True

        # Make sure we have at least one unique column
        self._enforce_unique_column(all_columns)

        # sort columns
        sorted_columns = self._sort_columns(all_columns)

        return sorted_columns

    def _enforce_unique_column(self, all_columns):
        # We must have at least one unique key. If no unique keys are enabled, then find the primary key and mark is as unique, default and enabled.
        if any([c.get('unique', False) for c in all_columns]):
            # fine, we have a unique column
            return
        for c in all_columns:
            if c.get('primary-key', False):
                c['unique'] = True
                c['default'] = True
                c['enabled'] = True

    def _column(self, column, in_use_columns):
        # a single attribute per column suffices to address a row
        attribute = {
            'name': str(column.key),
            'type': str(column.type).lower()
        }

        # create foreign key attribute
        if column.foreign_keys:
            attribute['foreign-key'] = self._foreign_key(column.foreign_keys)

        result = {'attributes': [attribute]}

        # set in-use attribute
        if column.name in in_use_columns:
            result['in-use'] = True

        # get constraints
        constraints = self._get_constraints(column)

        # set primary key and unique attributes
        for constraint in constraints:
            if constraint.__class__ == PrimaryKeyConstraint:
                result['primary-key'] = True
            elif self._is_applicable_unique_constraint(constraint):
                result['unique'] = True

        # set default attribute
        default = not result.get('primary-key', False) and not attribute['name'] in ODOO_SYSTEM_COLUMNS and result.get('in-use', False)
        if default:
            result['default'] = True
            result['enabled'] = True

        return result

    def _get_in_use_columns(self, table):
        # create list of column names
        column_names = [c.name for c in table.columns]
        # create column expressions
        expr = select(*[func.bool_or(column.isnot(None)) for column in table.columns])
        # execute query
        self._cr.execute(str(expr))
        result = self._cr.fetchone()
        # zip and filter non-null columns
        result = [c[0] for c in zip(column_names, result) if c[1]]
        # return list
        return result

    def _get_constraints(self, column):
        constraints = [constraint for constraint in column.table.constraints if column in list(constraint.columns)]
        return constraints

    def _foreign_key(self, foreign_keys):
        if len(foreign_keys) != 1:
            raise ValueError(f"Expected 1 foreign key, found {len(foreign_keys)}")
        foreign_key = list(foreign_keys)[0]
        table = foreign_key.column.table
        name = foreign_key.column.name
        return {'table': table.name, 'name': name, 'attributes': self._unique_columns(table)}

    def _unique_columns(self, table):
        # find applicable constraints
        unique_constraints = [c for c in table.constraints if self._is_applicable_unique_constraint(c)]
        primary_key_constraints = [c for c in table.constraints if c.__class__ == PrimaryKeyConstraint]

        if len(unique_constraints) > 0:
            # use unique columns from first unique constraint
            columns = unique_constraints[0].columns
        elif len(primary_key_constraints) > 0:
            # fall back on primary key columns
            columns = primary_key_constraints[0].columns
        else:
            columns = []

        return [self._unique_column(c) for c in columns]

    def _is_applicable_unique_constraint(self, c):
        # must be a unique constraint
        if c.__class__ != UniqueConstraint:
            return False

        # don't set jsonb columns as unique, because the resulting dict is not hashable
        if any([isinstance(column.type, JSONB) for column in c.columns]):
            return False

        # else it's applicable
        return True

    def _unique_column(self, column):
        attribute = {'name': column.name, 'type': str(column.type).lower()}
        if column.foreign_keys:
            attribute['foreign-key'] = self._foreign_key(column.foreign_keys)

        return attribute

    def _sort_columns(self, columns):
        # get primary key columns
        primary_key_columns = [c for c in columns if c.get('primary-key', False)]

        # get unique columns, exclude primary keys
        unique_columns = [c for c in columns if not c.get('primary-key', False) and c.get('unique', False)]

        # get other columns
        other_columns = [c for c in columns if not c.get('primary-key', False) and not c.get('unique', False)]

        return sorted(primary_key_columns, key=lambda e: e['attributes'][0]['name']) \
            + sorted(unique_columns, key=lambda e: e['attributes'][0]['name']) \
            + sorted(other_columns, key=lambda e: e['attributes'][0]['name'])



from setuptools import setup, find_packages

setup(
    name='stimula',
    version='0.0.12',
    description='Toolset library for Simple Table Mapping Language (STML)',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    url='https://github.com/stmlio/stimula',
    author='Romke Jonker',
    author_email='romke@rnadesign.net',
    packages=find_packages(),
    install_requires=[
        'pyjwt',
        'sqlalchemy',
        'pandas>=2.1.3',
        'numpy>=1.22.4',
        'cryptography>=3.4.8,<4.0',
    ],
    classifiers=[
        'Programming Language :: Python :: 3',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.6',
    entry_points={
        'console_scripts': [
            'stimula = stimula.cli.cli:main',
        ],
    },
)



"""
This class merges two mapping headers into a single mapping. The purpose is to let the user submit a mapping header and then
present him/her with a merged mapping header that contains the requested columns and the default columns in the correct order.

The requirements for the merged mapping are:
    - requested columns appear in order as requested
    - requested columns replace default columns, matched by attribute names
    - default columns appear in their own order, 'as much as possible'.
    - only requested columns are enabled


Author: Romke Jonker
Email: romke@rnadesign.net
"""


class HeaderMerger:
    def merge(self, default_mapping, requested_mapping):

        # verify both mappings have the same table
        assert default_mapping['table'] == requested_mapping['table']

        # disable all default columns by removing 'in-use' attribute
        for column in default_mapping['columns']:
            column.pop('enabled', None)

        # map default columns by column key
        default_columns_map = {self._create_column_key(column): column for column in default_mapping['columns']}

        # map requested columns by column key, but only for columns that have attributes
        requested_columns_map = {self._create_column_key(column): column for column in requested_mapping['columns'] if 'attributes' in column}

        # merge requested columns into default columns
        result_columns_map = self._merge_maps(default_columns_map, requested_columns_map, self._merge_column)

        # convert result column map to list
        result_columns = list(result_columns_map.values())

        result = {
            'table': default_mapping['table'],
            'columns': result_columns
        }

        return result

    def _create_column_key(self, column):
        #     join attribute names
        return ':'.join([attribute['name'] for attribute in column['attributes']])

    def _merge_column(self, default_column, requested_column):
        # return requested column, but keep 'in-use' attribute from default column
        if default_column.get('in-use', False):
            requested_column['in-use'] = True

        # also keep the 'primary-key' attribute from default column
        if default_column.get('primary-key', False):
            requested_column['primary-key'] = True

        return requested_column

    def _merge_maps(self, a, b, f=lambda x, y: y):
        '''
        Sorts two dictionaries by their 'key' attribute.
        Resulting dictionary has:
        - all keys from a and b
        - values from b have precedence over values from a
        - order in b is preserved
        - order in a is preserved for the longest possible tail
        '''
        # get keys from b that do not exist in a

        # sort keys in a by their position in b
        b_keys = (list(b.keys()))
        i = len(b_keys) - 1
        c_keys = []
        for key in reversed(a.keys()):
            if key in c_keys:
                # already visited
                pass
            elif not key in b_keys:
                # not in b, simply add to c
                c_keys.insert(0, key)
            else:
                while i >= 0 and key != b_keys[i]:
                    # must still add, but keep items from b in order
                    c_keys.insert(0, b_keys[i])
                    i -= 1
                if i >= 0:
                    # found key in b, add to c
                    c_keys.insert(0, key)
                    i -= 1
        # get remaining keys from b
        remainders = b_keys[:i + 1]
        # insert remainders at the front of c_keys
        c_keys = remainders + c_keys

        # merge dictionaries, using lambda f to merge values that appear in both dictionaries
        merged = {**a, **b}
        for key in a.keys() & b.keys():
            merged[key] = f(a[key], b[key])

        # sort by c_keys
        result = {k: merged[k] for k in c_keys}

        return result
